#ifndef PQCLEAN_SNTRUP761_AVX2_CRYPTO_VERIFY_1039_H
#define PQCLEAN_SNTRUP761_AVX2_CRYPTO_VERIFY_1039_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_AVX2_crypto_verify_1039_BYTES 1039

int PQCLEAN_SNTRUP761_AVX2_crypto_verify_1039(const unsigned char *x, const unsigned char *y);
#endif
