#ifndef PQCLEAN_SNTRUP761_AVX2_CRYPTO_DECODE_761X4591_H
#define PQCLEAN_SNTRUP761_AVX2_CRYPTO_DECODE_761X4591_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_AVX2_crypto_decode_761x4591_STRBYTES 1158
#define PQCLEAN_SNTRUP761_AVX2_crypto_decode_761x4591_ITEMS 761
#define PQCLEAN_SNTRUP761_AVX2_crypto_decode_761x4591_ITEMBYTES 2

void PQCLEAN_SNTRUP761_AVX2_crypto_decode_761x4591(void *v, const unsigned char *s);
#endif
