#ifndef PQCLEAN_SNTRUP761_AVX2_CRYPTO_CORE_MULT3SNTRUP761_H
#define PQCLEAN_SNTRUP761_AVX2_CRYPTO_CORE_MULT3SNTRUP761_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_AVX2_crypto_core_mult3sntrup761_OUTPUTBYTES 761
#define PQCLEAN_SNTRUP761_AVX2_crypto_core_mult3sntrup761_INPUTBYTES 761
#define PQCLEAN_SNTRUP761_AVX2_crypto_core_mult3sntrup761_KEYBYTES 761
#define PQCLEAN_SNTRUP761_AVX2_crypto_core_mult3sntrup761_CONSTBYTES 0

int PQCLEAN_SNTRUP761_AVX2_crypto_core_mult3sntrup761(unsigned char *outbytes, const unsigned char *inbytes, const unsigned char *kbytes);
#endif
