#ifndef PQCLEAN_MCELIECE348864_AVX2_NAMESPACE_H
#define PQCLEAN_MCELIECE348864_AVX2_NAMESPACE_H

#define CRYPTO_NAMESPACE(fun) PQCLEAN_MCELIECE348864_AVX2_ ## fun
#define _CRYPTO_NAMESPACE(fun) _PQCLEAN_MCELIECE348864_AVX2_ ## fun

#endif
