#ifndef PQCLEAN_MCELIECE8192128F_CLEAN_CRYPTO_KEM_H
#define PQCLEAN_MCELIECE8192128F_CLEAN_CRYPTO_KEM_H

#define crypto_kem_keypair CRYPTO_NAMESPACE(crypto_kem_keypair)
#define crypto_kem_enc CRYPTO_NAMESPACE(crypto_kem_enc)
#define crypto_kem_dec CRYPTO_NAMESPACE(crypto_kem_dec)

#endif
