#ifndef PQCLEAN_SNTRUP761_AVX2_CRYPTO_DECODE_INT16_H
#define PQCLEAN_SNTRUP761_AVX2_CRYPTO_DECODE_INT16_H

#include <stdint.h>
#define crypto_core_multsntrup1277_STRBYTES 2
#define crypto_core_multsntrup1277_ITEMS 1
#define crypto_core_multsntrup1277_ITEMBYTES 2
void PQCLEAN_SNTRUP761_AVX2_crypto_decode_int16(void *x, const unsigned char *s);
#endif
