#ifndef PQCLEAN_SNTRUP761_AVX2_CRYPTO_DECODE_761XINT16_H
#define PQCLEAN_SNTRUP761_AVX2_CRYPTO_DECODE_761XINT16_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_AVX2_crypto_decode_761xint16_STRBYTES 1522
#define PQCLEAN_SNTRUP761_AVX2_crypto_decode_761xint16_ITEMS 761
#define PQCLEAN_SNTRUP761_AVX2_crypto_decode_761xint16_ITEMBYTES 2

void PQCLEAN_SNTRUP761_AVX2_crypto_decode_761xint16(void *v, const unsigned char *s);
#endif
