#ifndef PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_761X1531_H
#define PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_761X1531_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x1531_STRBYTES 1007
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x1531_ITEMS 761
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x1531_ITEMBYTES 2

void PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x1531(unsigned char *out, const void *v);
#endif
