#ifndef PQCLEAN_SNTRUP761_AVX2_CRYPTO_ENCODE_761X1531ROUND_H
#define PQCLEAN_SNTRUP761_AVX2_CRYPTO_ENCODE_761X1531ROUND_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_AVX2_crypto_encode_761x1531round_STRBYTES 1007
#define PQCLEAN_SNTRUP761_AVX2_crypto_encode_761x1531round_ITEMS 761
#define PQCLEAN_SNTRUP761_AVX2_crypto_encode_761x1531round_ITEMBYTES 2

void PQCLEAN_SNTRUP761_AVX2_crypto_encode_761x1531round(unsigned char *out, const void *v);
#endif
