#ifndef PQCLEAN_MCELIECE460896F_CLEAN_NAMESPACE_H
#define PQCLEAN_MCELIECE460896F_CLEAN_NAMESPACE_H

#define CRYPTO_NAMESPACE(fun) PQCLEAN_MCELIECE460896F_CLEAN_ ## fun
#define _CRYPTO_NAMESPACE(fun) _PQCLEAN_MCELIECE460896F_CLEAN_ ## fun

#endif
