#ifndef PQCLEAN_SNTRUP761_AVX2_CRYPTO_DECODE_761X1531_H
#define PQCLEAN_SNTRUP761_AVX2_CRYPTO_DECODE_761X1531_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_AVX2_crypto_decode_761x1531_STRBYTES 1007
#define PQCLEAN_SNTRUP761_AVX2_crypto_decode_761x1531_ITEMS 761
#define PQCLEAN_SNTRUP761_AVX2_crypto_decode_761x1531_ITEMBYTES 2

void PQCLEAN_SNTRUP761_AVX2_crypto_decode_761x1531(void *v, const unsigned char *s);
#endif
