#ifndef PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_INT16_H
#define PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_INT16_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_int16_STRBYTES 2
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_int16_ITEMS 1
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_int16_ITEMBYTES 2

void PQCLEAN_SNTRUP761_CLEAN_crypto_encode_int16(unsigned char *s, const void *x);
#endif
